@@ -4,6 +4,7 @@ import datetime
 import logging
 import time
 import uuid
+from collections import defaultdict
 from email.header import Header
 from email.mime.multipart import MIMEMultipart
 from email.mime.text import MIMEText
@@ -20,12 +21,12 @@ from genshi.core import END, START, Attrs, QName
 from genshi.template import MarkupTemplate, TextTemplate
 from sql import Literal
 from sql.aggregate import Count
-from sql.functions import Position, Substring
 
 from trytond.config import config
 from trytond.i18n import gettext
 from trytond.model import (
-    EvalEnvironment, ModelSQL, ModelView, Unique, Workflow, dualmethod, fields)
+    EvalEnvironment, Index, ModelSQL, ModelView, Unique, Workflow, dualmethod,
+    fields)
 from trytond.pool import Pool
 from trytond.pyson import Eval, If, PYSONDecoder
 from trytond.report import Report
@@ -61,7 +62,7 @@ class Scenario(Workflow, ModelSQL, ModelView):
     "Marketing Scenario"
     __name__ = 'marketing.automation.scenario'
 
-    name = fields.Char("Name")
+    name = fields.Char("Name", translate=True)
     model = fields.Selection('get_models', "Model", required=True)
     domain = fields.Char(
         "Domain", required=True,
@@ -72,6 +73,9 @@ class Scenario(Workflow, ModelSQL, ModelView):
         fields.Integer("Records"), 'get_record_count')
     record_count_blocked = fields.Function(
         fields.Integer("Records Blocked"), 'get_record_count')
+    unsubscribable = fields.Boolean(
+        "Unsubscribable",
+        help="If checked parties are also unsubscribed from the scenario.")
     state = fields.Selection([
             ('draft', "Draft"),
             ('running', "Running"),
@@ -81,6 +85,11 @@ class Scenario(Workflow, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t, (t.state, Index.Equality()),
+                where=t.state.in_(['draft', 'running'])))
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'stopped'),
@@ -108,6 +117,10 @@ class Scenario(Workflow, ModelSQL, ModelView):
     def default_domain(cls):
         return '[]'
 
+    @classmethod
+    def default_unsubscribable(cls):
+        return False
+
     @classmethod
     def get_models(cls):
         pool = Pool()
@@ -208,10 +221,14 @@ class Scenario(Workflow, ModelSQL, ModelView):
 
         for scenario in scenarios:
             Model = pool.get(scenario.model)
-            sql_int = scenario.__class__.id.sql_type().base
             record = Record.__table__()
             cursor = Transaction().connection.cursor()
             domain = PYSONDecoder({}).decode(scenario.domain)
+            domain = [
+                domain,
+                ('marketing_party.marketing_scenario_unsubscribed',
+                    'not where', [('id', '=', scenario.id)]),
+                ]
             try:
                 query = Model.search(domain, query=True, order=[])
             except Exception:
@@ -221,10 +238,7 @@ class Scenario(Workflow, ModelSQL, ModelView):
                 continue
             cursor.execute(*(
                     query - record.select(
-                        Substring(
-                            record.record,
-                            Position(',', record.record) + Literal(1)
-                            ).cast(sql_int),
+                        Record.record.sql_id(record.record, Model),
                         where=record.scenario == scenario.id)))
 
             records = []
@@ -250,7 +264,7 @@ class Activity(ModelSQL, ModelView):
     "Marketing Activity"
     __name__ = 'marketing.automation.activity'
 
-    name = fields.Char("Name", required=True)
+    name = fields.Char("Name", translate=True, required=True)
     parent = fields.Reference(
         "Parent", [
             ('marketing.automation.scenario', "Scenario"),
@@ -506,28 +520,12 @@ class Activity(ModelSQL, ModelView):
                 for child in self.children])
 
     def _email_recipient(self, record):
-        pool = Pool()
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            Sale = pool.get('sale.sale')
-        except KeyError:
-            Sale = None
-
-        def get_party_email(party):
-            contact = party.contact_mechanism_get('email')
-            if contact and contact.email:
-                return _formataddr(
-                    contact.name or party.rec_name,
-                    contact.email)
-            return None
-
-        if Party and isinstance(record, Party):
-            return get_party_email(record)
-        elif Sale and isinstance(record, Sale):
-            return get_party_email(record.party)
+        party = record.marketing_party
+        contact = party.contact_mechanism_get('email')
+        if contact and contact.email:
+            return _formataddr(
+                contact.name or party.rec_name,
+                contact.email)
 
     def execute_send_email(
             self, record_activity, smtpd_datamanager=None, **kwargs):
@@ -569,7 +567,7 @@ class Activity(ModelSQL, ModelView):
             for kind, data, pos in stream:
                 if kind is START:
                     tag, attrs = data
-                    if tag == 'a' and attrs.get('href'):
+                    if tag == 'a':
                         href = attrs.get('href')
                         attrs -= 'href'
                         if href.startswith('unsubscribe'):
@@ -638,8 +636,12 @@ class Record(ModelSQL, ModelView):
         required=True, ondelete='CASCADE')
     record = fields.Reference(
         "Record", selection='get_models', required=True)
-    blocked = fields.Boolean("Blocked")
-    uuid = fields.Char("UUID", readonly=True)
+    blocked = fields.Boolean(
+        "Blocked",
+        states={
+            'readonly': ~Eval('blocked', False),
+            })
+    uuid = fields.Char("UUID", readonly=True, strip=False)
 
     @classmethod
     def __setup__(cls):
@@ -652,6 +654,11 @@ class Record(ModelSQL, ModelView):
             ('uuid_unique', Unique(t, t.uuid),
                 'marketing_automation.msg_record_uuid_unique'),
             ]
+        cls._buttons.update({
+                'block': {
+                    'invisible': Eval('blocked', False),
+                    },
+                })
 
     @classmethod
     def default_uuid(cls):
@@ -683,27 +690,26 @@ class Record(ModelSQL, ModelView):
 
     @property
     def language(self):
-        pool = Pool()
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            Sale = pool.get('sale.sale')
-        except KeyError:
-            Sale = None
-
-        if Party and isinstance(self.record, Party):
-            if self.record.lang:
-                return self.record.lang.code
-        elif Sale and isinstance(self.record, Sale):
-            if self.record.party.lang:
-                return self.record.party.lang.code
+        lang = self.record.marketing_party.lang
+        if lang:
+            return lang.code
 
     @dualmethod
+    @ModelView.button
     def block(cls, records):
+        pool = Pool()
+        Party = pool.get('party.party')
+
         cls.write(records, {'blocked': True})
 
+        parties = defaultdict(set)
+        for record in records:
+            if record.scenario.unsubscribable:
+                parties[record.record.marketing_party].add(record.scenario.id)
+        Party.write(*sum((
+                    ([p], {'marketing_scenario_unsubscribed': [('add', s)]})
+                    for p, s in parties.items()), ()))
+
     def get_rec_name(self, name):
         if self.record:
             return self.record.rec_name
@@ -762,6 +768,11 @@ class RecordActivity(Workflow, ModelSQL, ModelView):
             ('activity_record_unique', Unique(t, t.activity, t.record),
                 'marketing_automation.msg_activity_record_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_(['waiting'])))
         cls._transitions |= set((
                 ('waiting', 'done'),
                 ('waiting', 'cancelled'),
